EXPORTS
	OpenAdapter10
	OpenAdapter10_2
	D3DKMTAcquireKeyedMutex
	D3DKMTCloseAdapter
	D3DKMTConfigureSharedResource
	D3DKMTCreateAllocation
	D3DKMTCreateAllocation2
	D3DKMTCreateContext
	D3DKMTCreateDevice
	D3DKMTCreateKeyedMutex
	D3DKMTCreateSynchronizationObject
	D3DKMTCreateSynchronizationObject2
	D3DKMTDestroyAllocation
	D3DKMTDestroyContext
	D3DKMTDestroyDevice
	D3DKMTDestroyKeyedMutex
	D3DKMTDestroySynchronizationObject
	D3DKMTEscape
	D3DKMTGetContextSchedulingPriority
	;D3DKMTGetDeviceSchedulingPriority
	D3DKMTGetDeviceState
	D3DKMTGetDisplayModeList
	D3DKMTGetMultisampleMethodList
	D3DKMTGetRuntimeData
	D3DKMTGetSharedPrimaryHandle
	D3DKMTLock
	D3DKMTOpenAdapterFromDeviceName
	D3DKMTOpenAdapterFromGdiDisplayName
	D3DKMTOpenKeyedMutex
	D3DKMTOpenResource
	D3DKMTOpenResource2
	D3DKMTOpenSynchronizationObject
	D3DKMTPresent
	D3DKMTQueryAdapterInfo
	D3DKMTQueryAllocationResidency
	D3DKMTQueryResourceInfo
	D3DKMTReleaseKeyedMutex
	D3DKMTRender
	D3DKMTSetAllocationPriority
	D3DKMTSetContextSchedulingPriority
	;D3DKMTSetDeviceSchedulingPriority
	D3DKMTSetDisplayMode
	D3DKMTSetDisplayPrivateDriverFormat
	D3DKMTSetGammaRamp
	D3DKMTSetVidPnSourceOwner
	D3DKMTSetVidPnSourceOwner1
	D3DKMTSignalSynchronizationObject
	D3DKMTSignalSynchronizationObject2
	D3DKMTUnlock
	D3DKMTWaitForSynchronizationObject
	D3DKMTWaitForSynchronizationObject2
	D3DKMTWaitForVerticalBlankEvent
